#!/usr/bin/env node
'use strict';

var path = require('path');
var fs = require('fs');
var commander = require('commander');
var stdio_js = require('@modelcontextprotocol/sdk/server/stdio.js');
var mcp_js = require('@modelcontextprotocol/sdk/server/mcp.js');
var zod = require('zod');
var axios = require('axios');
var _ = require('lodash');
var morgan = require('morgan');
var express = require('express');
var sse_js = require('@modelcontextprotocol/sdk/server/sse.js');

/**
 * 日志工具
 *
 * 提供统一的日志输出接口，支持不同级别的日志，包括彩色输出
 */

// 日志级别枚举
var LogLevel; (function (LogLevel) {
  const DEBUG = 0; LogLevel[LogLevel["DEBUG"] = DEBUG] = "DEBUG";
  const INFO = 1; LogLevel[LogLevel["INFO"] = INFO] = "INFO";
  const WARN = 2; LogLevel[LogLevel["WARN"] = WARN] = "WARN";
  const ERROR = 3; LogLevel[LogLevel["ERROR"] = ERROR] = "ERROR";
  const SUCCESS = 4; LogLevel[LogLevel["SUCCESS"] = SUCCESS] = "SUCCESS";
  const NONE = 5; LogLevel[LogLevel["NONE"] = NONE] = "NONE"; // 用于完全禁用日志
})(LogLevel || (LogLevel = {}));

// ANSI 颜色代码
const colors = {
  reset: '\x1b[0m',
  debug: '\x1b[36m', // 青色
  info: '\x1b[34m', // 蓝色
  warn: '\x1b[33m', // 黄色
  error: '\x1b[31m', // 红色
  success: '\x1b[32m', // 绿色
};

// 默认日志级别
let currentLogLevel = LogLevel.INFO;

/**
 * 基础日志函数
 * @param level 日志级别
 * @param message 日志消息
 * @param showPrefix 是否显示前缀，默认为 true
 */
function log(level, message, showPrefix = true) {
  if (level < currentLogLevel) {
    return;
  }

  let prefix = '';
  let color = '';

  switch (level) {
    case LogLevel.DEBUG: {
      prefix = '[DEBUG]';
      color = colors.debug;
      break;
    }
    case LogLevel.WARN: {
      prefix = '[WARN]';
      color = colors.warn;
      break;
    }
    case LogLevel.ERROR: {
      prefix = '[ERROR]';
      color = colors.error;
      break;
    }
    case LogLevel.SUCCESS: {
      prefix = '[SUCCESS]';
      color = colors.success;
      break;
    }
    case LogLevel.INFO:
    default: {
      prefix = '[INFO]';
      color = colors.info;
      break;
    }
  }

  // 格式化日志前缀
  const formattedPrefix = showPrefix ? `${color}${prefix}${colors.reset} ` : '';

  // 输出日志
  console.log(`${formattedPrefix}${message}`);
}

/**
 * 导出的日志工具
 */
const logger = {
  /**
   * 调试日志
   * @param message 日志消息
   * @param showPrefix 是否显示前缀，默认为 true
   */
  debug(message, showPrefix = true) {
    log(LogLevel.DEBUG, message, showPrefix);
  },

  /**
   * 信息日志
   * @param message 日志消息
   * @param showPrefix 是否显示前缀，默认为 true
   */
  info(message, showPrefix = true) {
    log(LogLevel.INFO, message, showPrefix);
  },

  /**
   * 警告日志
   * @param message 日志消息
   * @param showPrefix 是否显示前缀，默认为 true
   */
  warn(message, showPrefix = true) {
    log(LogLevel.WARN, message, showPrefix);
  },

  /**
   * 错误日志
   * @param message 日志消息
   * @param showPrefix 是否显示前缀，默认为 true
   */
  error(message, showPrefix = true) {
    log(LogLevel.ERROR, message, showPrefix);
  },

  /**
   * 成功日志
   * @param message 日志消息
   * @param showPrefix 是否显示前缀，默认为 true
   */
  success(message, showPrefix = true) {
    log(LogLevel.SUCCESS, message, showPrefix);
  },
};

const schema$1 = zod.z.object({
  keyword: zod.z.string().min(1),
  size: zod.z.number().min(1).max(100).default(20),
  from: zod.z.number().min(0).default(0),
  quality: zod.z.number().min(0).max(1).default(0.65),
  popularity: zod.z.number().min(0).max(1).default(0.98),
  maintenance: zod.z.number().min(0).max(1).default(0.5),
  registry: zod.z.string().url().default('https://registry.npmjs.org'),
});

const formatResults = (results) => JSON.stringify(results);

const searchNpmPackages = {
  name: 'searchNpmPackages',
  description:
    'Search for NPM packages with advanced filters. Use this tool when you need to find suitable NPM packages for a specific use case, compare package popularity and maintenance scores, or discover alternatives to existing packages. You can customize the search by adjusting quality, popularity, and maintenance thresholds, and specify a custom registry if needed.',
  schema: schema$1,
  execute: async ({
    keyword,
    size,
    from,
    quality,
    popularity,
    maintenance,
    registry,
  }) => {
    logger.info(
      `Start searching NPM packages, keyword: "${keyword}", size: ${size}, from: ${from}`,
    );
    logger.debug(
      `Search parameters - quality: ${quality}, popularity: ${popularity}, maintenance: ${maintenance}, registry: ${registry}`,
    );

    try {
      const searchEndpoint = `${registry.replace(/\/$/, '')}/-/v1/search`;
      logger.debug(`Search endpoint URL: ${searchEndpoint}`);

      logger.debug('Sending API request');
      const response = await axios.get(searchEndpoint, {
        params: {
          text: keyword,
          size,
          from,
          quality,
          popularity,
          maintenance,
        },
      });

      logger.debug(
        `Successfully received response, found ${response.data.objects.length} packages`,
      );

      const results = response.data.objects.map(obj => ({
        package: {
          name: obj.package.name,
          version: obj.package.version,
          description: obj.package.description,
          keywords: obj.package.keywords,
          author: obj.package.author,
          publisher: obj.package.publisher,
          date: obj.package.date,
        },
        score: obj.score,
        searchScore: obj.searchScore,
      }));

      logger.debug('Formatting search results');
      const content = formatResults(results);
      logger.success(`Successfully completed NPM package search: "${keyword}"`);
      return { content: [{ type: 'text', text: content }] };
    } catch (error) {
      logger.error(
        `Failed to search NPM packages: ${error instanceof Error ? error.message : 'Unknown error'}`,
      );
      const errorMessage = `# Error Searching NPM Packages\n\n${error instanceof Error ? error.message : 'Unknown error'}`;
      return { content: [{ type: 'text', text: errorMessage }] };
    }
  },
};

function _optionalChain(ops) { let lastAccessLHS = undefined; let value = ops[0]; let i = 1; while (i < ops.length) { const op = ops[i]; const fn = ops[i + 1]; i += 2; if ((op === 'optionalAccess' || op === 'optionalCall') && value == null) { return undefined; } if (op === 'access' || op === 'optionalAccess') { lastAccessLHS = value; value = fn(value); } else if (op === 'call' || op === 'optionalCall') { value = fn((...args) => value.call(lastAccessLHS, ...args)); lastAccessLHS = undefined; } } return value; }

const schema = zod.z.object({
  packageNames: zod.z.array(zod.z.string()).min(1),
  registry: zod.z.string().url().default('https://registry.npmjs.org'),
});

const fetchPackage = async (
  packageName,
  registry,
) => {
  try {
    const response = await axios.get(`${registry}/${packageName}`);

    return response.data ;
  } catch (error) {
    logger.error(
      `[npm] Failed to fetch package ${packageName}: ${(error ).message}`,
    );
    throw error;
  }
};

const simply = (result) => {
  const simplyVersion = versionDescription =>
    _.pick(versionDescription, [
      'name',
      'version',
      'description',
      'author',
      'bugs',
    ]);
  const distTags = result['dist-tags'];
  const tags = Object.keys(distTags);
  // version 内容过多了，需要简化一下
  const tagVersions = Object.keys(result.versions)
    .filter(v => tags.includes(v))
    .reduce(
      (acc, v) => ({
        ...acc,
        [v]: simplyVersion(result.versions[v]),
      }),
      {},
    );

  return {
    ..._.pick(result, [
      'name',
      'description',
      'author',
      'bugs',
      'license',
      'homepage',
      'keywords',
      'repository',
      'contributors',
      'maintainers',
      'distTags',
    ]),
    versions: tagVersions,
    latest: simplyVersion(result.versions[distTags.latest || distTags[0]]),
    distTags,
  };
};



const format = (result) => JSON.stringify(result);

const fetchNpmPackages = {
  name: 'fetchNpmPackages',
  description:
    'Fetch multiple npm packages information. Use this tool when you need detailed metadata about specific npm packages, such as version history, dependencies, maintainers, and other package.json information. This is ideal for comparing package details, checking package compatibility, verifying package authenticity, or gathering comprehensive information about packages before installing them. You can specify a custom registry if needed.',
  schema,
  execute: async ({ packageNames, registry }) => {
    logger.info(
      `[npm] Starting to fetch package info for ${packageNames.length} packages from registry: ${registry}`,
    );

    // 使用 Promise.all 并发请求，但限制并发数为 5
    const batchSize = 5;
    const results = [];

    for (let i = 0; i < packageNames.length; i += batchSize) {
      const batch = packageNames.slice(i, i + batchSize);
      logger.info(
        `[npm] Processing batch ${Math.floor(i / batchSize) + 1}/${Math.ceil(packageNames.length / batchSize)}, containing ${batch.length} packages`,
      );

      const batchResults = await Promise.all(
        batch.map(async packageName => {
          try {
            const result = await fetchPackage(packageName, registry);
            const simplyResult = simply(result);
            logger.success(
              `[npm] Successfully fetched package info: ${packageName}, latest version: ${simplyResult.latest.version}`,
            );
            return simplyResult;
          } catch (e) {
            if (e instanceof axios.AxiosError && _optionalChain([e, 'access', _2 => _2.response, 'optionalAccess', _3 => _3.status]) === 404) {
              logger.error(`[npm] Package not found: ${packageName}`);
              return { success: false, message: 'Package not found' };
            }
            logger.error(
              `[npm] Failed to fetch package info: ${packageName}, error: ${(e ).message}`,
            );
            return { success: false, message: (e ).message };
          }
        }),
      );

      logger.info(`[npm] Batch ${Math.floor(i / batchSize) + 1} completed`);
      results.push(...batchResults);
    }

    logger.success(
      `[npm] All package info fetched, total results: ${results.length}`,
    );
    const content = format(results);
    return { content: [{ type: 'text', text: content }] };
  },
};

const SERVER_NAME = 'npm-mcp';

const SERVER_DESCRIPTION = `NPM MCP Server - A tool for LLMs to interact with npm registry.

Capabilities:
1. Search for npm packages with advanced filtering by quality, popularity, and maintenance scores
2. Fetch detailed metadata for specific npm packages including version history, dependencies, and author information
3. Support for custom npm registries, including private registries and mirrors like npmmirror.com

Use cases:
- When you need to find suitable packages for a specific use case or development requirement
- When you need to compare similar packages based on quality metrics and popularity
- When you need to analyze package dependencies and compatibility before installation
- When you need to verify package authenticity and maintenance status
- When you need to retrieve comprehensive metadata from specific npm packages`;

const createServer = () => {
  logger.debug(`Creating MCP server: ${SERVER_NAME}`);
  const server = new mcp_js.McpServer({
    name: SERVER_NAME,
    description: SERVER_DESCRIPTION,
    version: '1.0.0',
  });
  const tools = [fetchNpmPackages, searchNpmPackages];

  tools.forEach(tool => {
    server.tool(tool.name, tool.description, tool.schema.shape, tool.execute);
  });

  return server;
};

/**
 * Start MCP service in CLI mode
 * @param componentsDir Components documentation directory
 */
const startAction = async () => {
  try {
    const server = createServer();
    const transport = new stdio_js.StdioServerTransport();
    logger.debug('Connecting to StdioServerTransport');
    server.connect(transport);

    logger.success('MDC MCP service started (CLI mode)');

    logger.info('Use Ctrl+C to exit the service');
    // Keep the process running until manually exited by user
    return new Promise(resolve => {
      process.on('SIGINT', () => {
        resolve();
        logger.info('Closing MCP service...');
        process.exit(0);
      });
    });
  } catch (error) {
    logger.error(
      `Failed to start MCP service: ${error instanceof Error ? error.message : String(error)}`,
    );
    process.exit(1);
  }
};

const createApp = () => {
  const app = express();
  app.use(morgan('tiny'));
  // app.use(express.json());
  // app.use(express.urlencoded({ extended: true }));
  logger.debug('Express application created successfully');
  return app;
};

/**
 * Start HTTP server
 * @param port Port number
 * @param componentsDir Components documentation directory
 */
const serveAction = (port) => {
  const server = createServer();
  try {
    const app = createApp();

    let transport;
    app.get('/sse', async (req, res) => {
      logger.debug(`Received SSE connection request: ${req.ip}`);
      transport = new sse_js.SSEServerTransport('/messages', res);
      await server.connect(transport);
    });

    app.post('/messages', async (req, res) => {
      // Note: to support multiple simultaneous connections, these messages will
      // need to be routed to a specific matching transport. (This logic isn't
      // implemented here, for simplicity.)
      logger.debug(`Received message request: ${req.ip}`);
      await transport.handlePostMessage(req, res);
    });

    app.listen(port, () => {
      logger.success(`MCP server started, listening on port: ${port}`);
    });
  } catch (error) {
    logger.error(
      `Failed to start MCP server: ${error instanceof Error ? error.message : String(error)}`,
    );
    process.exit(1);
  }
};

const main = () => {
  // Create command line program
  const program = new commander.Command();
  const programName = 'npm-mcp-server';

  // Set version and description
  program
    .name(programName)
    .description('MCP server for npm')
    .version(
      JSON.parse(
        fs.readFileSync(path.resolve(__dirname, '../package.json'), 'utf-8'),
      ).version,
    );

  // Add serve command - Start HTTP server
  program
    .command('serve')
    .description('Start the MCP HTTP server')
    .option('-p, --port <port>', 'Port to listen on', '3000')
    .action(async options => {
      const port = parseInt(options.port, 10);
      logger.info(`Starting HTTP service, port: ${port}`);
      await serveAction(port);
      logger.info(`You can access MCP service at http://localhost:${port}/sse`);
    });

  program
    .command('start')
    .description('Start the MCP server in terminal')
    .action(async () => {
      await startAction();
    });

  const SPLIT_LINE = `${'='.repeat(35)}`;
  // Parse command line arguments
  logger.success(
    `${SPLIT_LINE}\n🎉 Welcome to npm MCP CLI! 🎉\n${SPLIT_LINE}`,
    false,
  );
  program.parse(process.argv);

  // If no command is provided, show help information
  if (!process.argv.slice(2).length) {
    program.outputHelp();
  }
};

// Execute main function
main();
